/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattisbvf
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __PIKA_IMAGE_COLORMAP_H__
#define __PIKA_IMAGE_COLORMAP_H__


#define PIKA_IMAGE_COLORMAP_SIZE 768


void           pika_image_colormap_init            (PikaImage     *image);
void           pika_image_colormap_dispose         (PikaImage     *image);
void           pika_image_colormap_free            (PikaImage     *image);

void           pika_image_colormap_update_formats  (PikaImage     *image);

const Babl   * pika_image_colormap_get_rgb_format  (PikaImage     *image);
const Babl   * pika_image_colormap_get_rgba_format (PikaImage     *image);

PikaPalette  * pika_image_get_colormap_palette     (PikaImage     *image);
void           pika_image_set_colormap_palette     (PikaImage     *image,
                                                    PikaPalette   *palette,
                                                    gboolean       push_undo);

guchar *       pika_image_get_colormap             (PikaImage     *image);
gint           pika_image_get_colormap_size        (PikaImage     *image);
void           pika_image_set_colormap             (PikaImage     *image,
                                                    const guchar  *colormap,
                                                    gint           n_colors,
                                                    gboolean       push_undo);
void           pika_image_unset_colormap           (PikaImage     *image,
                                                    gboolean       push_undo);

gboolean       pika_image_colormap_is_index_used   (PikaImage     *image,
                                                    gint           color_index);

void           pika_image_get_colormap_entry       (PikaImage     *image,
                                                    gint           color_index,
                                                    PikaRGB       *color);
void           pika_image_set_colormap_entry       (PikaImage     *image,
                                                    gint           color_index,
                                                    const PikaRGB *color,
                                                    gboolean       push_undo);

void           pika_image_add_colormap_entry       (PikaImage     *image,
                                                    const PikaRGB *color);
gboolean       pika_image_delete_colormap_entry    (PikaImage     *image,
                                                    gint           color_index,
                                                    gboolean       push_undo);


#endif /* __PIKA_IMAGE_COLORMAP_H__ */
